# -*- coding: utf-8 -*-

import dataiku
import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score, accuracy_score, f1_score, precision_score, recall_score, hamming_loss, log_loss
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error, median_absolute_error, explained_variance_score
from sklearn.base import BaseEstimator
import re
import logging
import RuleGeneratorUtils
from TreeExtractor import treeExtract


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,  # avoid getting log from 3rd party module
                    format='RuleGeneration plugin %(levelname)s - %(message)s')


class RuleGenerator(BaseEstimator):

    """
    The main object of the RuleGeneration package.

    Public methods (conform the norm of a sklearn's Estimator):
    - fit
    - predict
    - predict_proba

    Private methods (in order of appearance in the fit method):
    - _create_df_from_doctor_input
    - _eval_rules
        - _create_train_set_from_rules
        - _extract_feature_from_rules
        - _check_stopping_criteria
        - _verbalize

    """

    def __init__(self, params=None):

        assert params is not None, "RuleGeneratorParams instance is None"
        params.check()
        self.params = params

    def fit(self, X, y):
        """
        For each tree estimator in our Random Forest, extract the rules to get to the leaf nodes. 
        We will then use these rules as the new features and apply the feature extractor algorithm.
        """

        # recreate the dataframe from X and y, we need their columns name for
        # the algorithm
        train_df = self._create_df_from_doctor_input(X, y.values)

        # cache this value for the PREDICT phase
        self.params.mean_target_value = np.mean(y)

        # STEP 1: Extract trees from a random forest estimator, then extract
        # rules from trees and use them as new features
        random_forest_estimator = self.params.random_forest_estimator
        raw_rules_df = random_forest_estimator.extract_rules_from_trees(
            train_df, self.params.feature_list, self.params.target)
        rules_df = self._eval_rules(raw_rules_df, train_df)
        # new_train_df is the new updated train df with rule columns, rules_df
        # is the dataframe storing the rules (as rows)
        new_train_df = self._create_train_set_from_rules(rules_df, train_df)

        logger.info(
            "Initally fitting the regularized method and keeping only the features which appears in the selected rules.")
        # before selecting features, save the original feature list length
        new_feature_list_len = len(self.params.feature_list)
        logger.info("Initial number of features: {}".format(
            new_feature_list_len))
        logger.info("Initial number of rules: {}".format(len(rules_df)))

        # STEP 2: filter out rules using the ruleRegularizer object
        rule_regularizer = self.params.rule_regularizer

        features_to_consider = list(
            set(new_train_df.columns) - set([self.params.target, self.params.ids_col]))

        rules_list = rule_regularizer.get_rules_list(
            new_train_df, features_to_consider, self.params.target)
        rules_df = rule_regularizer.get_rules_df(rules_list, rules_df)

        feature_list = self._extract_feature_from_rules(rules_df=rules_df)
        self.params.feature_list = feature_list

        # if alpha is too strong, there will be no rules left after
        # lasso/logreg
        if len(self.params.feature_list) == 0:
            raise ValueError(
                'Your regularization is too strong. Please decrease your alpha parameter.')

        logger.info("Starting the iterative rule selection process with {} features".format(
            len(self.params.feature_list)))
        logger.info("Starting the iterative rule selection process with {} rules".format(
            len(rules_df)))

        stop_flag = False
        performance_drop_percentage = 0

        # repeat until stable, which means no feature is removed from the
        # regularized method or the stopping conditions are satisfied
        while(len(self.params.feature_list) < new_feature_list_len):

            # cache the old rules in case the new rules dont satisfy the
            # criteria
            old_rules_df = rules_df

            new_feature_list_len = len(self.params.feature_list)
            logger.info("Number of features in use: {}".format(
                len(self.params.feature_list)))

            raw_rules_df = random_forest_estimator.extract_rules_from_trees(
                train_df, self.params.feature_list, self.params.target)
            rules_df = self._eval_rules(
                rules_df=raw_rules_df, train_df=train_df)
            new_train_df = self._create_train_set_from_rules(
                rules_df=rules_df, old_train_df=train_df)

            logger.info('Checking stopping criteria ...')
            initial_performance = random_forest_estimator.params.performance_cache[
                0]
            current_performance = random_forest_estimator.params.performance_cache[
                -1]
            performance_drop_percentage = 100 * \
                (initial_performance - current_performance) / \
                float(abs(initial_performance))
            stop_flag = self._check_stopping_criteria(
                rules_df, performance_drop_percentage)

            if stop_flag:
                # revert back to the previous rules
                rules_df = old_rules_df
                break
            else:
                features_to_consider = list(
                    set(new_train_df.columns) - set([self.params.target, self.params.ids_col]))
                rules_list = rule_regularizer.get_rules_list(
                    new_train_df, features_to_consider, self.params.target)
                rules_df = rule_regularizer.get_rules_df(rules_list, rules_df)

                feature_list = self._extract_feature_from_rules(
                    rules_df=rules_df)
                self.params.feature_list = feature_list

            logger.info("Number of features left: {}".format(
                len(self.params.feature_list)))
            logger.info("Number of rules left: {}".format(len(rules_df)))

        # if the process stops "naturally", check the final rules_df
        if not stop_flag and len(self.params.feature_list) == new_feature_list_len:
            logger.info('Final check')
            _ = self._check_stopping_criteria(
                rules_df, performance_drop_percentage, raise_error=True)

        logger.info("The itertive rule selection process ended with {} rules applied on {} features.".format(
            len(rules_df), len(self.params.feature_list)))

        # before caching, we will add a column that stores the value of
        # mean_target_value, target and ml_type
        rules_df['mean_target_value'] = self.params.mean_target_value
        rules_df['target'] = self.params.target
        rules_df['ml_type'] = self.params.random_forest_estimator.params.ml_type

        self.params.rules_df = rules_df

    def _create_df_from_doctor_input(self, X, y):
        """
        Retrieve the list of features from the list of rules. 

        Parameters
        ----------

        X : nD-array

        y : 1D array 


        Returns
        -------

        train_df: dataframe
        """

        self.classes_ = list(set(y))
        self.n_classes = len(self.classes_)

        y = y.reshape([-1, 1])
        d = np.concatenate((X, y), axis=1)

        train_df = pd.DataFrame(
            d, columns=self.params.feature_list + [self.params.target])
        train_df = train_df.dropna(how='any')

        # add the ids col
        train_df[self.params.ids_col] = range(len(train_df))

        return train_df

    def _eval_rules(self, rules_df, train_df):
        """
        Parameters
        ----------

        rules_df : dataframe
                The extracted_tree dataframe, each row is the information of a node in the tree.


        train_df : dataframe
                Dataframe with features and target which will be used to evaluate the rules.


        Returns
        -------
        _ : dataframe
                The rules dataframe of leaf nodes only with metrics.

        """

        if self.params.random_forest_estimator.params.ml_type == 'clf':

            rules_df["support_precision"] = rules_df.rule.apply(lambda one_rule: RuleGeneratorUtils.evaluateRuleClf(
                self.params.target, one_rule, train_df[self.params.feature_list + [self.params.target]]))
            rules_df["population_percentage"] = rules_df.support_precision.apply(
                lambda support_precision: support_precision[0])
            rules_df["precision"] = rules_df.support_precision.apply(
                lambda support_precision: support_precision[1])
            del rules_df["support_precision"]

            return rules_df.loc[:, ['rule', 'rule_verbal', 'population_percentage', 'precision', 'impurity']]

        elif self.params.random_forest_estimator.params.ml_type == 'reg':

            rules_df["support_mean_var"] = rules_df.rule.apply(lambda one_rule: RuleGeneratorUtils.evaluateRuleReg(
                self.params.target, one_rule, train_df[self.params.feature_list + [self.params.target]]))
            rules_df["population_percentage"] = rules_df.support_mean_var.apply(
                lambda support_mean_var: support_mean_var[0])
            rules_df["target_mean"] = rules_df.support_mean_var.apply(
                lambda support_mean_var: support_mean_var[1])
            rules_df["target_variance"] = rules_df.support_mean_var.apply(
                lambda support_mean_var: support_mean_var[2])
            del rules_df["support_mean_var"]

            return rules_df.loc[:, ['rule', 'rule_verbal', 'population_percentage', 'target_mean', 'target_variance']]

    # THIS FUNCTION TAKES QUITE A LOT OF TIME !
    def _create_train_set_from_rules(self, rules_df, old_train_df):
        """
        Create a new trainset with the rules as features. 
        Those will be binary columns: 1 if the current row satistfies that rule and vice versa.

        Parameters
        ----------

        rules_df : dataframe
                The extracted_tree dataframe.


        old_train_df : dataframe
                Base dataframe with features and target.


        Returns
        -------

        dataframe: new train set with each rule as a new binary column.
        """

        df = old_train_df[[self.params.ids_col, self.params.target]].copy()

        for rule, rule_verbal in rules_df[['rule', 'rule_verbal']].values.tolist():
            ids_good = list(old_train_df.query(rule)[self.params.ids_col])
            df_temp = pd.DataFrame(
                {self.params.ids_col: ids_good, rule_verbal: len(ids_good) * [1]})
            # each rule becomes a binary col of whether or not that specific
            # observation satisfies the rule
            df = pd.merge(df, df_temp, how='left')

        # rows that don't satisfy a rule are filled with zero
        new_train_df = df.fillna(0)

        return new_train_df

    def _extract_feature_from_rules(self, rules_df):
        """
        Retrieve the list of features from the list of rules. 

        Parameters
        ----------

        rules_df : dataframe
                The extracted_tree dataframe.


        Returns
        -------

        new_feature_list: list of string
        """

        rules_list = rules_df['rule_verbal'].values.tolist()

        raw_feature_list = rules_list
        raw_feature_list = [x.replace('between', '') for x in raw_feature_list]
        raw_feature_list = [x.replace('above', '') for x in raw_feature_list]
        raw_feature_list = [x.replace('below', '') for x in raw_feature_list]
        raw_feature_list = [x.split(' & ') for x in raw_feature_list]

        selected_feature_list = []
        for x in raw_feature_list:  # retrieve final feature list
            l_ = [y.split(' ') for y in x]
            l_ = [y[0] for y in l_]
            l_ = list(set(l_))

            selected_feature_list = selected_feature_list + l_

        new_feature_list = list(set(selected_feature_list))

        return new_feature_list

    def _check_stopping_criteria(self, rules_df, performance_drop_percentage, raise_error=False):
        """
        Check the new rules_df to see if it violates the expected conditions. If yes, send back a flag to stop the algorithm.
        List of conditions to respect:

        * Performance drop is less than a threshold defined by user.
        * Max support is greater than a threshold defnied by user.


        Parameters
        ----------

        rules_df: dataframe
                The extracted_tree dataframe, each row is the information of a node in the tree.

        performance_drop_percentage: float

        raise_error: boolean

        Returns
        -------

        stop_flag: Boolean
                Binary flag to tell the algorithm whether or not it should continue the optimisation process.

        """

        max_support = 100 * np.max(rules_df['population_percentage'])
        stop_flag = False

        if max_support < self.params.max_support_threshold:
            if raise_error:
                raise ValueError("OPTIMISATION CONSTRAINT NOT SATISFIED: the minimum value of max population_percentage ({0:.2f}%) is below the allowed threshold ({1:.2f}%). You should decrease the number of clauses in a rule.".format(
                    max_support, self.params.max_support_threshold))

            # if max support criteria fails twice in a row, stop the algorithm
            if self.params.failed_max_support_criteria_count > 2:
                stop_flag = True
                logger.warn("OPTIMISATION PROCESS STOPS: the minimum value of max population_percentage ({0:.2f}%) is below the allowed threshold ({1:.2f}%) twice in a row. You should decrease the number of clauses in a rule.".format(
                    max_support, self.params.max_support_threshold))
                self.params.failed_max_support_criteria_count = 0
            else:
                self.params.failed_max_support_criteria_count += 1
        else:
            self.params.failed_max_support_criteria_count = 0

        if performance_drop_percentage > self.params.performance_drop_threshold:
            if raise_error:
                raise ValueError("OPTIMISATION CONSTRAINT NOT SATISFIED: the performance drop percentage ({0:.2f}%) is above the allowed threshold ({1:.2f}%). You should try to decrease the regularization term and/or increase the number of clauses in a rule".format(
                    performance_drop_percentage, self.params.performance_drop_threshold))
            # if performance drop criteria failes twice in a row, stop the
            # algorithm
            if self.params.failed_performance_drop_criteria_count > 2:
                stop_flag = True
                logger.warn("OPTIMISATION PROCESS STOPS: the performance drop percentage ({0:.2f}%) is above the allowed threshold ({1:.2f}%) twice in a row. You should try to decrease the regularization term and/or increase the number of clauses in a rule".format(
                    performance_drop_percentage, self.params.performance_drop_threshold))
                self.params.failed_performance_drop_criteria_count = 0
            else:
                self.params.failed_performance_drop_criteria_count += 1
        else:
            self.params.failed_performance_drop_criteria_count = 0

        return stop_flag

    def predict(self, X):
        """    
        Scoring new dataset using the best set of rules.

        - Compute the "precision" (percentage of target = 1) of all rules.
        - For an observation, take all compatible rules and compute the mean of the precision of each rule. 
        It gives you the probability of your observation.
        - If the observation doesn't follow any rules of your set, put the global percentage of target = 1 in 
        your train set as the probability.
        """

        df = pd.DataFrame(X, columns=['feature_{}'.format(
            i) for i in xrange(len(self.params.raw_feature_list))])

        l = len(df)
        # retrieve only the selected features.
        column_list = list(self.params.feature_list)

        # in case we are evaluate a labelled dataset, add the target column to
        # the output dataset
        if self.params.target in df.columns:
            column_list.append(self.params.target)

        # check if the scoring df has all the necessary columns
        if len(set(column_list) - set(df.columns.tolist())) != 0:
            raise ValueError(
                'The scoring dataset does not have all the needed features.')

        df[self.params.ids_col] = range(len(df))
        i = 0

        for rule in self.params.rules_df.iterrows():

            i = i + 1
            ids_good = list(df.query(rule[1]['rule'])[self.params.ids_col])
            l = len(ids_good)

            # create a df with rows as the returned ids and the corresponding
            # value of their precision/target mean
            if self.params.random_forest_estimator.params.ml_type == 'clf':  # classification
                temp_res = pd.DataFrame({self.params.ids_col: ids_good, 'value_rule_' + str(
                    i): l * [rule[1]['precision']]})  # proba of label 1
            else:  # regression
                temp_res = pd.DataFrame({self.params.ids_col: ids_good, 'value_rule_' + str(
                    i): l * [rule[1]['target_mean']]})  # mean target value

            df = pd.merge(df, temp_res, how='left')

        var_p = ['value_rule_' + str(x) for x in range(1, i + 1)]

        if self.params.random_forest_estimator.params.ml_type == 'clf':

            df['proba_'] = df[var_p].apply(lambda x: np.mean(
                x), axis=1)  # average of all rule's proba
            # for rows that dont satisfy any rules, fill with mean target value
            df['proba_'] = df['proba_'].fillna(self.params.mean_target_value)
            # TODO: this should be a parameter ?
            df['prediction'] = df['proba_'] >= 0.5
            column_list.extend(['proba_', 'prediction'])

        else:

            df['prediction'] = df[var_p].apply(lambda x: np.mean(
                x), axis=1)  # average of all rule's target value
            df['prediction'] = df['prediction'].fillna(
                self.params.mean_target_value)
            column_list.append('prediction')

        column_list = list(set(column_list))
        scored_df = df[column_list]

        return np.array(scored_df['prediction'])

    def predict_proba(self, X):
        """    
        Scoring new dataset using the best set of rules.

        - Compute the "precision" (percentage of target = 1) of all rules.
        - For an observation, take all compatible rules and compute the mean of the precision of each rule. 
        It gives you the probability of your observation.
        - If the observation doesn't follow any rules of your set, put the global percentage of target = 1 in 
        your train set as the probability.
        """

        df = pd.DataFrame(X, columns=['feature_{}'.format(
            i) for i in xrange(len(self.params.raw_feature_list))])

        l = len(df)
        # retrieve only the selected features.
        column_list = list(self.params.feature_list)

        # in case we are evaluate a labelled dataset, add the target column to
        # the output dataset
        if self.params.target in df.columns:
            column_list.append(self.params.target)

        # check if the scoring df has all the necessary columns
        if len(set(column_list) - set(df.columns.tolist())) != 0:
            raise ValueError(
                'The scoring dataset does not have all the needed features.')

        df[self.params.ids_col] = range(len(df))
        i = 0

        for rule in self.params.rules_df.iterrows():

            i = i + 1
            ids_good = list(df.query(rule[1]['rule'])[self.params.ids_col])
            l = len(ids_good)

            # create a df with rows as the returned ids and the corresponding
            # value of their precision/target mean
            if self.params.random_forest_estimator.params.ml_type == 'clf':  # classification
                temp_res = pd.DataFrame({self.params.ids_col: ids_good, 'value_rule_' + str(
                    i): l * [rule[1]['precision']]})  # proba of label 1
            else:  # regression
                temp_res = pd.DataFrame({self.params.ids_col: ids_good, 'value_rule_' + str(
                    i): l * [rule[1]['target_mean']]})  # mean target value

            df = pd.merge(df, temp_res, how='left')

        var_p = ['value_rule_' + str(x) for x in range(1, i + 1)]

        if self.params.random_forest_estimator.params.ml_type == 'clf':

            df['proba_'] = df[var_p].apply(lambda x: np.mean(
                x), axis=1)  # average of all rule's proba
            # for rows that dont satisfy any rules, fill with mean target value
            df['proba_'] = df['proba_'].fillna(self.params.mean_target_value)
            # TODO: this should be a parameter ?
            df['prediction'] = df['proba_'] >= 0.5
            column_list.extend(['proba_', 'prediction'])

        else:

            df['prediction'] = df[var_p].apply(lambda x: np.mean(
                x), axis=1)  # average of all rule's target value
            df['prediction'] = df['prediction'].fillna(
                self.params.mean_target_value)
            column_list.append('prediction')

        column_list = list(set(column_list))
        scored_df = df[column_list]

        proba_1 = df['proba_'].values
        proba_0 = 1 - proba_1

        return np.stack((proba_0, proba_1), axis=-1)

    def retrieve_rules(self):
        """
        Return the final rules dataframe.

        Return
        ------

        translated_rules_df : dataframe: 
                The self.rules_df attribute, sorted by support value.

        """

        translated_rules_df = self.params.rules_df.sort_values(
            'population_percentage', ascending=False).copy()
        pattern = re.compile(
            r'\b(' + '|'.join(self.params.feature_dict.keys()) + r')\b')

        translated_rules_df['rule'] = translated_rules_df['rule'].apply(
            lambda y: pattern.sub(lambda x: self.params.feature_dict[x.group()], y))
        translated_rules_df['rule_verbal'] = translated_rules_df[
            'rule'].apply(self._verbalize)

        return translated_rules_df

    def _verbalize(self, rule):
        """
        Sincere sorry to whoever have to read this code ...

        Parameters
        ----------

        rule : string
                For example: "Embarked_value_S <= 0.5 & Pclass_value_3 > 0.5" 


        Returns
        -------

        string, verbolised rule.


        """

        listR = rule.split(" & ")
        d_inf = {}
        d_sup = {}

        c_inf = {}
        c_sup = {}

        l_inf = []
        l_sup = []

        for i in listR:

            if 'dummy:' in i:
                if "<=" in i:
                    f, tr = i.split(" <= ")
                    f1, f2, f3 = f.split(':')
                    c_inf[f] = f3
                    l_inf.append(f2)

                elif ">" in i:
                    f, tr = i.split(" > ")
                    f1, f2, f3 = f.split(':')
                    c_sup[f] = f3
                    l_sup.append(f2)
            else:
                if "<=" in i:
                    f, tr = i.split(" <= ")
                    d_inf[f] = tr

                elif ">" in i:
                    f, tr = i.split(" > ")
                    d_sup[f] = tr

        s = ""

        for f in sorted(list(set(d_inf.keys() + d_sup.keys()))):

            if f in d_inf and f in d_sup:
                s += f + " between " + d_sup[f] + " and " + d_inf[f]
            elif f in d_inf:
                s += f + " below " + d_inf[f]
            elif f in d_sup:
                s += f + " above " + d_sup[f]
            s += " & "

        for f in sorted(list(set(c_inf.keys() + c_sup.keys()))):

            f1, f2, f3 = f.split(':')

            if f in c_sup:
                s += f2 + " is " + c_sup[f]
                s += " & "

            elif f in c_inf:
                if f2 not in l_sup:
                    s += f2 + " is not " + c_inf[f]
                    s += " & "

        return s[:-3]
