# -*- coding: utf-8 -*-

import dataiku
from dataiku import pandasutils as pdu
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
import logging


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,  # avoid getting log from 3rd party module
                    format='RuleGeneration plugin %(levelname)s - %(message)s')


class RuleGeneratorParams(BaseEstimator):

    def __init__(self,
                 raw_feature_list=[],
                 target='',
                 random_forest_estimator=None,
                 rule_regularizer=None,
                 ids_col='NEW_id_',
                 max_support_threshold=60,
                 performance_drop_threshold=5
                 ):
        """
        Attributes
        ----------

        ml_type: string
                Regression or Classification.
        feature_list: list of string
                List of features chosen by user at the beginning.
        target: string
                Name of the column defining the target.
        ids_col: string
                Name of the column defining the ids of each row.
        rules_df: dataframe
                The dataframe of nodes and rules (as rows), will be set in the FIT phase.
        rules_list: list of string
                The list of rules, will be set in the FIT phase.        
        mean_target_value: float
                Mean value of target in the train set, will be set in FIT phase.
        """

        assert random_forest_estimator is not None, "random_forest_estimator instance is None"
        assert rule_regularizer is not None, "rule_regularizer instance is None"

        self.raw_feature_list = raw_feature_list
        self.target = target
        self.ids_col = ids_col
        self.random_forest_estimator = random_forest_estimator
        self.rule_regularizer = rule_regularizer
        self.max_support_threshold = max_support_threshold
        self.performance_drop_threshold = performance_drop_threshold

        # encoding features'name so that we don't have problem later on when
        # querying rules using pd.query(...)
        self.feature_dict = {'feature_{}'.format(i): x for (
            i, x) in enumerate(self.raw_feature_list)}
        self.feature_list = ['feature_{}'.format(
            i) for i in xrange(len(self.raw_feature_list))]

        self.rules_df = []
        self.mean_target_value = 0
        self.failed_max_support_criteria_count = 0
        self.failed_performance_drop_criteria_count = 0

    @property
    def raw_feature_list(self):
        return self._raw_feature_list

    @raw_feature_list.setter
    def raw_feature_list(self, value):
        if not isinstance(value, list) or len(value) == 0:
            raise ValueError('Feature list must be a non-empty list.')
        self._raw_feature_list = value

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        if not isinstance(value, str) or value == '':
            raise ValueError('Target column name must be defined.')
        self._target = value

    @property
    def max_support_threshold(self):
        return self._max_support_threshold

    @max_support_threshold.setter
    def max_support_threshold(self, value):
        if value < 0 or value > 100:
            raise ValueError(
                'max_support_threshold must be between 0% and 100%.')
        self._max_support_threshold = value

    @property
    def performance_drop_threshold(self):
        return self._performance_drop_threshold

    @performance_drop_threshold.setter
    def performance_drop_threshold(self, value):
        if value < 0:
            raise ValueError(
                'performance_drop_threshold must be between 0% and 100%.')
        self._performance_drop_threshold = value

    @property
    def ids_col(self):
        return self._ids_col

    @ids_col.setter
    def ids_col(self, value):
        self._ids_col = value

    @property
    def failed_max_support_criteria_count(self):
        return self._failed_max_support_criteria_count

    @failed_max_support_criteria_count.setter
    def failed_max_support_criteria_count(self, value):
        self._failed_max_support_criteria_count = value

    @property
    def failed_performance_drop_criteria_count(self):
        return self._failed_performance_drop_criteria_count

    @failed_performance_drop_criteria_count.setter
    def failed_performance_drop_criteria_count(self, value):
        self._failed_performance_drop_criteria_count = value

    @property
    def random_forest_estimator(self):
        return self._random_forest_estimator

    @random_forest_estimator.setter
    def random_forest_estimator(self, value):
        self._random_forest_estimator = value

    @property
    def rule_regularizer(self):
        return self._rule_regularizer

    @rule_regularizer.setter
    def rule_regularizer(self, value):
        self._rule_regularizer = value

    @property
    def rules_df(self):
        return self._rules_df

    @rules_df.setter
    def rules_df(self, value):
        self._rules_df = value

    @property
    def mean_target_value(self):
        return self._mean_target_value

    @mean_target_value.setter
    def mean_target_value(self, value):
        self._mean_target_value = value

    def check(self):

        self.random_forest_estimator.params.check()
        self.rule_regularizer.params.check()
