# -*- coding: utf-8 -*-

import dataiku
from dataiku import pandasutils as pdu
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier as RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor as RandomForestRegressor
from sklearn.model_selection import GridSearchCV
import logging
from sklearn.base import BaseEstimator
import logging
from TreeExtractor import treeExtract


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,  # avoid getting log from 3rd party module
                    format='RuleGeneration plugin %(levelname)s - %(message)s')


class RuleGeneratorRandomForest(BaseEstimator):

    def __init__(self, params=None):

        assert params is not None, "RuleGeneratorRandomForestParams instance is None."
        params.check()
        self.params = params

    def extract_rules_from_trees(self, df, feature_list, target):
        """
        Parameters
        ----------
        df : dataframe              

        feature_list : list of string

        target : string

        Returns
        -------
        rules_df: dataframe

        """
        clf = self. _create_grid_search_cv()
        # this will do GridSeach as clf is a GriSearchCV instance
        clf.fit(df[feature_list], df[target])
        self.params.performance_cache.append(clf.best_score_)

        first_iteration = True
        rules_df = pd.DataFrame()
        tree_estimator_list = clf.best_estimator_.estimators_

        for tree_estimator in tree_estimator_list:

            # step 1: retrieve the dictionary containing the computed
            # extracted_tree dataframe.
            extracted_tree_dict = self._extract_info_from_tree_node(
                tree_estimator, feature_list)

            # step 2: retrieve rules on the leaf's nodes
            raw_node_rules_df = extracted_tree_dict["extracted_tree"]
            raw_leaf_rules_df = raw_node_rules_df.loc[
                raw_node_rules_df['children_left'] == -1]
            temp_rules_df = raw_leaf_rules_df.loc[:, ['rule', 'rule_verbal']]

            if first_iteration:
                rules_df = temp_rules_df
                first_iteration = False

            else:  # from the second iteration of this loop, start appending rule columns to the new train df
                rules_df = pd.concat([rules_df, temp_rules_df])

        rules_df = rules_df.drop_duplicates()

        return rules_df

    def _create_grid_search_cv(self):
        """
        Returns
        -------
        clf: GridSearchCV instance
        """

        if self.params.ml_type == 'clf':
            chosen_clf = RandomForestClassifier(random_state=1407, n_jobs=-1)
            logger.info(
                "The trees will be extracted from a RandomForestClassifier.")
        else:
            chosen_clf = RandomForestRegressor(random_state=1407, n_jobs=-1)
            logger.info(
                "The trees will be extracted from a RandomForestRegressor.")

        clf = GridSearchCV(estimator=chosen_clf,
                           param_grid=self.params.param_grid,
                           cv=2,
                           scoring=self.params.optimisation_metric,
                           verbose=1,
                           n_jobs=-1)

        return clf

    def _extract_info_from_tree_node(self, tree_estimator, feature_list):
        """
        Apply rules extractor to a tree.   

        - Step 1 : instantiate the treeExtract object.

        - Step 2 : update the extrtacted_tree dataframe with information concerning the rules' path.

        In extracted_tree dataframe, each row is a node in the tree, columns are information about their path, their parent, ...


        Parameters
        ----------
        tree estimator : object   
                Sklearn's decision tree estimator.              

        feature_list : list of string
                Input dataframe without target

        Returns
        -------
        A dictionary containing the ``extracted_tree`` dataframe. 
        """

        extract = treeExtract(tree_estimator, columns=feature_list)
        extract.fit(ml_type=self.params.ml_type)

        return {"extracted_tree": extract.extracted_tree}
